#!/usr/bin/env python

from distutils.core import setup
import os
import glob
import sys

name = "python-mytools"
version = "0.0.8"

setup(name=name,
	version=version,
	description="Utilities to integrate MySQL into Python applications",
	long_description="mytools is a set of tools designed to integrate MySQL with standard Python applications, including fast indexing, catalogs and private database clusters.\n\npython-mytools is part of the Search services project.",
	author="Manuel Amador (Rudd-O)",
	author_email="rudd-o@amautacorp.com",
	url="http://www.amautacorp.com/staff/Rudd-O/projects/search-services/",
	packages=["mytools"],
	)
